import { ResourcesKeys } from "./keys";
import { resourceInformation } from "./resources";
import { Store } from "./store";

const methods: Array<
  (resourceKey: ResourcesKeys, store: Store) => number | undefined
> = [
  getRawChemicalsLimit,
  getBrainwashingMaterialLimit,
  getSubjectsLimit,
  getBrainwashingMachineLimit,
  getBlankDroneLimit,
  getDroneLimitExceptServiceWhenOneRemaining,
];

export function getResourceLimit(
  resourceKey: ResourcesKeys,
  store: Store
): number | undefined {
  let total: number | undefined = undefined;
  for (const method of methods) {
    const result = method(resourceKey, store);
    if (result === undefined) continue;
    if (total === undefined) total = result;
    else total = Math.min(total, result);
  }
  return total;
}

export function getTotalDroneLimitFromServiceDrones(store: Store) {
  const serviceDrones = store.resources.serviceDrone?.quantity || 0;
  if (serviceDrones < 1) return 10;
  return 20 * serviceDrones;
}

function getRawChemicalsLimit(
  resourceKey: ResourcesKeys,
  store: Store
): number | undefined {
  if (resourceKey !== "rawChemicals") return undefined;
  const currentLevel = store.research.rawChemicalsStorage?.level || 0;
  return Math.pow(2, currentLevel) * 50;
}

function getBrainwashingMaterialLimit(
  resourceKey: ResourcesKeys,
  store: Store
): number | undefined {
  if (resourceKey !== "brainwashingMaterials") return undefined;
  const currentLevel = store.research.brainwashingMaterialsStorage?.level || 0;
  return Math.pow(2, currentLevel) * 100;
}

function getSubjectsLimit(
  resourceKey: ResourcesKeys,
  store: Store
): number | undefined {
  if (resourceKey !== "unprocessedSubjects") return undefined;

  // Special case: Disable acquiring new subjects during the tutorial.
  if (
    store.storyEvents.includes("unlock-blank-drone") &&
    !store.storyEvents.includes("unlock-second-subject")
  ) {
    return 0;
  }

  // Special case: Disable acquiring new subjects when the whole planet has been gotten.
  if (store.world.freePopulation === 0) return 0;

  const currentLevel = store.research.holdingCells?.level || 0;
  if (currentLevel === 0) return 1;

  return Math.pow(2, currentLevel);
}

function getBrainwashingMachineLimit(
  resourceKey: ResourcesKeys,
  store: Store
): number | undefined {
  if (resourceKey !== "brainwashingChambers") return undefined;
  const maintenanceDrones = store.resources.maintenanceDrone?.quantity || 0;
  return 1 + maintenanceDrones;
}

function getBlankDroneLimit(
  resourceKey: ResourcesKeys,
  store: Store
): number | undefined {
  if (resourceKey !== "blankDrones") return undefined;
  const limit = getTotalDroneLimitFromServiceDrones(store);
  const totalDrones = store.world.dronePopulation;
  const remainingAvailableDrones = limit - totalDrones;
  const blankDrones = store.resources.blankDrones?.quantity || 0;
  const currentHoldingCellsLevel = store.research.holdingCells?.level || 0;
  const maxDrones =
    currentHoldingCellsLevel === 0
      ? 10
      : Math.pow(2, currentHoldingCellsLevel) + 10;
  return Math.min(remainingAvailableDrones + blankDrones, maxDrones);
}

function getDroneLimitExceptServiceWhenOneRemaining(
  resourceKey: ResourcesKeys,
  store: Store
): number | undefined {
  const resourceInfo = resourceInformation[resourceKey];
  if (!resourceInfo.tags.includes("drone")) return undefined;
  if (resourceKey === "blankDrones") return undefined;
  if (resourceKey === "serviceDrone") return undefined;
  const limit = getTotalDroneLimitFromServiceDrones(store);
  const totalDrones = store.world.dronePopulation;
  if (totalDrones < limit) return undefined; // We can still brainwash new drones. Go ahead.
  const blankDrones = store.resources.blankDrones?.quantity || 0;
  if (blankDrones > 1) return undefined; // We still have several blank drones left. Go ahead.
  // Prevent a softlock by disabling building any drone except service drones
  return 0;
}
