import { ResearchKeys } from "./keys";
import { Store, StoreTaskResearchGoal } from "./store";

export type ResearchInformation = {
  id: ResearchKeys;
  name: string;
  description: string;
  duration: number;
};

export function* getAvailableResearch(
  store: Store
): IterableIterator<ResearchInformation> {
  const runningResearches = store.tasks
    .map((t) => t.goal)
    .filter((g): g is StoreTaskResearchGoal => g.type === "research")
    .map((g) => g.research);

  const scienceDrones = store.resources.researcherDrones?.quantity || 0;
  const durationFactor = scienceDrones > 0 ? 1.0 / scienceDrones : 1.0;
  const speedUpgradeRatio = 2 / (1 + scienceDrones * scienceDrones * 0.3);
  const reducedSpeedUpgradeRatio = 3 / (1 + scienceDrones * 0.1);

  const minHours = 2;

  function isResearching(id: ResearchKeys): boolean {
    return runningResearches.includes(id);
  }

  function isUnlocked(id: ResearchKeys): boolean {
    return (store.research[id]?.level || 0) > 0;
  }

  function isResearchingOrUnlocked(id: ResearchKeys): boolean {
    return isResearching(id) || isUnlocked(id);
  }

  const gathererDroneCount = store.resources.gathererDrones?.quantity || 0;
  const crafterDroneCount = store.resources.crafterDrones?.quantity || 0;
  const serviceDroneCount = store.resources.serviceDrone?.quantity || 0;
  const maintenanceDroneCount = store.resources.maintenanceDrone?.quantity || 0;
  const retrieverDroneCount = store.resources.retrieverDrone?.quantity || 0;
  const builderDroneCount = store.resources.builderDrones?.quantity || 0;
  const converterDroneCount = store.resources.converterDrones?.quantity || 0;
  const researcherDroneCount = store.resources.researcherDrones?.quantity || 0;
  const overseerDroneCount = store.resources.overseerDrones?.quantity || 0;

  if (!isResearching("rawChemicalsStorage")) {
    const currentLevel = store.research.rawChemicalsStorage?.level || 0;
    if (store.world.maxRawChemicalsStored >= Math.pow(2, currentLevel) * 50) {
      yield {
        id: "rawChemicalsStorage",
        name: "Expand shelf storage",
        description:
          "Increase the maximum amount of raw chemicals that can be stored.",
        duration: Math.max(
          currentLevel === 0 ? 5 : currentLevel * 5 * durationFactor,
          minHours
        ),
      };
    }
  }

  if (!isResearching("brainwashingMaterialsStorage")) {
    const currentLevel =
      store.research.brainwashingMaterialsStorage?.level || 0;
    if (
      store.world.maxBrainwashingMaterialsStored >=
      Math.pow(2, currentLevel) * 100
    ) {
      yield {
        id: "brainwashingMaterialsStorage",
        name: "Expand refregirated storage",
        description:
          "Increase the maximum amount of brainwashing materials that can be stored.",
        duration: Math.max(
          currentLevel === 0 ? 5 : currentLevel * 5 * durationFactor,
          minHours
        ),
      };
    }
  }

  if (gathererDroneCount > 0 && !isResearching("rawChemicalsSpeed")) {
    const currentLevel = store.research.rawChemicalsSpeed?.level || 0;
    const maxLevel = Math.log(gathererDroneCount / speedUpgradeRatio);
    if (currentLevel < maxLevel) {
      if (currentLevel === 0) {
        yield {
          id: "rawChemicalsSpeed",
          name: "Raw Chemicals E-mail Alerts",
          description:
            "Set up an email alert for availability of raw chemicals, making the gatherer drones faster.",
          duration: Math.max(6 * durationFactor, minHours),
        };
      } else if (currentLevel === 1) {
        yield {
          id: "rawChemicalsSpeed",
          name: "Start producing raw chemicals in-house",
          description:
            "Obtain some chemical manufacturing equipment and teach the gatherer drones how to use it, increasing their production speed.",
          duration: Math.max(currentLevel * 6 * durationFactor, minHours),
        };
      } else {
        yield {
          id: "rawChemicalsSpeed",
          name: "Improve in-house raw chemicals production",
          description:
            "Improve the production equipment and gardens, making raw chemicals production even faster.",
          duration: Math.max(currentLevel * 6 * durationFactor, minHours),
        };
      }
    }
  }

  if (crafterDroneCount > 0 && !isResearching("brainwashingMaterialsSpeed")) {
    const currentLevel = store.research.brainwashingMaterialsSpeed?.level || 0;
    const maxLevel = Math.log(crafterDroneCount / speedUpgradeRatio);
    if (currentLevel < maxLevel) {
      yield {
        id: "brainwashingMaterialsSpeed",
        name: "Inproved crafting instructions",
        description:
          "Increase the crafting drone's speed for creating brainwashing material.",
        duration: Math.max(
          currentLevel === 0 ? 6 : currentLevel * 6 * durationFactor,
          minHours
        ),
      };
    }
  }

  if (retrieverDroneCount > 0 && !isResearching("retrieverSpeedup")) {
    const currentLevel = store.research.retrieverSpeedup?.level || 0;
    const maxLevel = Math.log(retrieverDroneCount / speedUpgradeRatio);
    const entries: Array<{ name: string; description: string }> = [
      {
        name: "Widen retrieval criteria",
        description:
          "Widen the criteria for retrieval from 'immediate threat of exposing the hive' to 'possible threat of exposing the hive', making Retriever Drones find new targets faster.",
      },
      {
        name: "Widen Retrieval Criteria",
        description:
          "Widen the criteria for retrieval from 'possible threat of exposing the hive' to 'possible threat of knowing about the hive', making Retriever Drones find new targets even faster.",
      },
      {
        name: "Widen retrieval criteria",
        description:
          "Widen the criteria for retrieval from 'possible threat of knowing about the hive' to 'possible threat of encountering drones', making Retriever Drones find new targets even faster.",
      },
      {
        name: "Widen Retrieval Criteria",
        description:
          "Widen the criteria for retrieval from 'possible threat of encountering drones' to 'possible threat of encountering potential drones', making Retriever Drones find new targets even faster.",
      },
      {
        name: "Retrieval Files",
        description:
          "Start pre-filing information about possible retrieval cases, making Retriever Drones even faster.",
      },
      {
        name: "Improved Retrieval Files",
        description:
          "Pre-file information for all possible cases related to retrieval, making Retriever Drones even faster.",
      },
      {
        name: "Retriever Safe Houses",
        description:
          "Train Retriever Drones to settle as common folks in controlled safe houses, making retrieval operations even faster.",
      },
      {
        name: "Retriever Tracking Training",
        description:
          "Train Retriever Drones for finding people who are trying to escape them, making them even faster.",
      },
      {
        name: "Pre-Retrieval",
        description:
          "Train some Retriever Drones into Pre-retrievers focused on preventing messy Retrievals, making them even faster.",
      },
      {
        name: "Trained Pre-Retrievers",
        description:
          "Train all of your Retriever Drones in both Retrieval and Pre-Retrieval, making them even faster.",
      },
    ];
    if (currentLevel < maxLevel) {
      if (currentLevel < entries.length) {
        yield {
          id: "retrieverSpeedup",
          duration: Math.max(6 * (currentLevel + 1) * durationFactor, minHours),
          ...entries[currentLevel],
        };
      } else {
        yield {
          id: "retrieverSpeedup",
          name: "Retrieval Efficiency Training",
          description:
            "Improve the training of Retriever Drones further, making them even faster.",
          duration: Math.max(6 * durationFactor * (currentLevel + 1), minHours),
        };
      }
    }
  }

  if (retrieverDroneCount > 0 && !isResearching("holdingCells")) {
    const currentLevel = store.research.holdingCells?.level || 0;
    if (currentLevel === 0) {
      yield {
        id: "holdingCells",
        name: "Holding cell",
        description:
          "Build a holding cell for guests that don't really want to be there. It's not really a cell, though. It has cushions and stuff.",
        duration: Math.max(durationFactor, minHours),
      };
    } else {
      if (Math.pow(2, currentLevel) < retrieverDroneCount) {
        yield {
          id: "holdingCells",
          name: "Extra holding cells",
          description:
            "Double your number of holding cells, allowing extra spots for unprocessed subjects.",
          duration: Math.max(durationFactor * (currentLevel + 1), minHours),
        };
      }
    }
  }

  if (converterDroneCount > 1 && !isResearching("conversionSpeedup")) {
    const currentLevel = store.research.conversionSpeedup?.level || 0;
    const maxLevel = Math.log(converterDroneCount / reducedSpeedUpgradeRatio);
    if (currentLevel < maxLevel) {
      yield {
        id: "conversionSpeedup",
        name: "Improve conversion speed",
        description:
          "Increase the speed at which your converter drones imprint blank drones.",
        duration: Math.max(
          currentLevel === 0 ? 6 : currentLevel * 6 * durationFactor,
          minHours
        ),
      };
    }
  }

  if (
    converterDroneCount > 0 &&
    maintenanceDroneCount >= 5 &&
    !isResearchingOrUnlocked("automaticConversionMaintenance")
  ) {
    yield {
      id: "automaticConversionMaintenance",
      name: "Train Converter Drones to create Maintenance Drones",
      description:
        "Training people to maintain machines is just as slow as building them yourself. You can automate this part.",
      duration: 8,
    };
  }

  if (
    converterDroneCount > 0 &&
    builderDroneCount >= 1 &&
    isUnlocked("automaticConversionMaintenance") &&
    !isResearchingOrUnlocked("automaticConversionBuilder")
  ) {
    yield {
      id: "automaticConversionBuilder",
      name: "Train Converter Drones to create Builder Drones",
      description:
        "More brainwashing machines = more good, so automating the builders too is a good idea! Probably!",
      duration: 16,
    };
  }

  if (
    converterDroneCount > 0 &&
    serviceDroneCount >= 5 &&
    !isResearchingOrUnlocked("automaticConversionService")
  ) {
    yield {
      id: "automaticConversionService",
      name: "Train Converter Drones to create Service Drones",
      description:
        "Ok, that technically means infinite growth, but it's not too dangerous. It's still linear. And you kinda want a linear increase of drones anyways.",
      duration: 8,
    };
  }

  if (
    converterDroneCount > 0 &&
    retrieverDroneCount >= 5 &&
    isUnlocked("automaticConversionService") &&
    !isResearchingOrUnlocked("automaticConversionRetriever")
  ) {
    yield {
      id: "automaticConversionRetriever",
      name: "Train Converter Drones to create Retriever Drones",
      description:
        "It's probably fine to make more retriever drones automatically? In theory, you don't need too many of these, sure. But they're useful to have around. And they're very nice to look at.",
      duration: 16,
    };
  }

  if (
    converterDroneCount > 0 &&
    researcherDroneCount >= 5 &&
    overseerDroneCount > 0 &&
    isUnlocked("automaticConversionBuilder") &&
    isUnlocked("automaticConversionRetriever") &&
    !isResearchingOrUnlocked("automaticConversionResearcher")
  ) {
    yield {
      id: "automaticConversionResearcher",
      name: "Train Converter Drones to create Researcher Drones",
      description:
        "These are very useful to do more research faster. You don't really risk anything just having more of them.",
      duration: 16,
    };
  }

  if (
    converterDroneCount >= 5 &&
    isUnlocked("automaticConversionResearcher") &&
    store.storyEvents.includes("unlock-converter-drone-conversion") &&
    !isResearchingOrUnlocked("automaticConversionConverter")
  ) {
    yield {
      id: "automaticConversionConverter",
      name: "Train Converter Drones to create Converter Drones",
      description:
        "You know that, in theory, it's dangerous to do that because of exponential growth and runaway effect and stuff. But, come on, it's so freakin' slow to train them all yourself. Would it really be so bad to just make them train each other?",
      duration: 16,
    };
  }

  if (
    overseerDroneCount >= 4 &&
    isUnlocked("automaticConversionConverter") &&
    store.storyEvents.includes("unlock-overseer-drone-conversion") &&
    !isResearchingOrUnlocked("automaticConversionOverseer")
  ) {
    yield {
      id: "automaticConversionOverseer",
      name: "Train Converter Drones to create Overseer Drones",
      description:
        "You already automated all the other drones, and it takes you two days each time you want a new Overseer Drone. Might as well automate it too.",
      duration: 32,
    };
  }

  if (overseerDroneCount > 0 && !isResearching("overseerEffectIncrease")) {
    const currentLevel = store.research.overseerEffectIncrease?.level || 0;
    const maxLevel = Math.log(overseerDroneCount / reducedSpeedUpgradeRatio);
    const entries: Array<{ name: string; description: string }> = [
      {
        name: "Overseer Drone Chatroom",
        description:
          "Create a chatroom for Overseer Drones, encouraging knowledge sharing and making their multiplicative power greater.",
      },
      {
        name: "Overseer Drone Templates",
        description:
          "Regularly share brainwashing templates created by groups of Overseer Drones among other Overseer Drones, making their multiplicative power even greater.",
      },
      {
        name: "Overseer Overseeing",
        description:
          "Assign Overseer Drones to improve each other's work as well as other drone's, making their multiplicative power even greater.",
      },
      {
        name: "Drone Assignment Criteria",
        description:
          "Allow Overseer Drones to require pre-existing shell characteristics for the more specialized Drone categories, making the multiplicative power of Overseer Drones even greater.",
      },
      {
        name: "Shared Thoughts Patterns",
        description:
          "Use brainwashing to copy interesting thoughts patterns between Overseer Drones, making their multiplicative power even greater.",
      },
      {
        name: "Regular Template Re-Imprinting",
        description:
          "Mandate regular re-imprinting of the updated templates on a set schedule, making the multiplicative power of Overseer Drones even greater.",
      },
      {
        name: "Drone Reassignment Schedules",
        description:
          "Mandate reassignment between drone categories on a set schedule, making the multiplicative power of Overseer Drones even greater.",
      },
    ];
    if (currentLevel < maxLevel) {
      if (currentLevel < entries.length) {
        yield {
          id: "overseerEffectIncrease",
          duration: Math.max(6 * (currentLevel + 1) * durationFactor, minHours),
          ...entries[currentLevel],
        };
      } else {
        yield {
          id: "overseerEffectIncrease",
          name: "Improve Overseer Efficiency",
          description:
            "Improve the efficiency of Overseer Drones further, making their multiplicative power even greater.",
          duration: Math.max(6 * (currentLevel + 1) * durationFactor, minHours),
        };
      }
    }
  }
}
