import { ResourcesKeys, ResourceTagKeys } from "./keys";

export type ResourceAcquisitionInformation = {
  text: string;
  cost?: Array<{
    key: ResourcesKeys;
    baseCost: number;
    additionalCost?: number;
  }>;
  sideEffect?:
    | "freeToCaptured"
    | "capturedToDrone"
    | "recomputeMaxRawChemicals"
    | "recomputeMaxBrainwashingMaterials";
};

export type AcquisitionResourceKey = Extract<
  ResourcesKeys,
  | "rawChemicals"
  | "unprocessedSubjects"
  | "brainwashingMaterials"
  | "blankDrones"
  | "brainwashingChambers"
>;

export type ResourceEffectInformation =
  | { type: "runAcquisitionFor"; key: AcquisitionResourceKey }
  | { type: "buildBrainwashingChambers" }
  | { type: "startResearchAvenue" }
  | { type: "convertBlankDrone" };

export type ResourceInformation = {
  name: string;
  tags: ResourceTagKeys[];
  fluffDescription: string;
  effectDescription: string;
  acquire: ResourceAcquisitionInformation;
  effect?: ResourceEffectInformation;
};

export const resourceInformation: {
  [key in ResourcesKeys]: ResourceInformation;
} = {
  rawChemicals: {
    name: "Raw Chemicals",
    tags: ["material"],
    fluffDescription:
      "A handful of pharmaceutical products with decently long names.",
    effectDescription: "Used to create brainwashing material.",
    acquire: {
      text: "Acquire",
      sideEffect: "recomputeMaxRawChemicals",
    },
  },
  brainwashingMaterials: {
    name: "Brainwashing Material",
    tags: ["material"],
    fluffDescription:
      "Everything you'd ever need for creating, changing, and upgrading a drone.",
    effectDescription:
      "Used to create blank drones, imprint drones, or build machines.",
    acquire: {
      text: "Craft",
      cost: [{ key: "rawChemicals", baseCost: 1 }],
      sideEffect: "recomputeMaxBrainwashingMaterials",
    },
  },
  unprocessedSubjects: {
    name: "Unprocessed Subject",
    tags: ["inhabitant"],
    fluffDescription: "Not yet compliant enough to do what is needed.",
    effectDescription: "Can be turned into a blank drone.",
    acquire: {
      text: "Acquire",
      sideEffect: "freeToCaptured",
    },
  },
  blankDrones: {
    name: "Blank Drone",
    tags: ["inhabitant", "drone"],
    fluffDescription:
      "This very efficient drone will do anything once it's been told what to do. Unfortunately, it's not yet been told what to do.",
    effectDescription: "Can be imprinted into other kinds of drones.",
    acquire: {
      text: "Brainwash",
      cost: [
        { key: "unprocessedSubjects", baseCost: 1 },
        { key: "brainwashingMaterials", baseCost: 1 },
      ],
      sideEffect: "capturedToDrone",
    },
  },
  gathererDrones: {
    name: "Gatherer Drone",
    tags: ["inhabitant", "drone"],
    fluffDescription:
      "Chemical components are hard to come by, at least for those who don't know where to look.",
    effectDescription: "Automatically acquires raw chemicals.",
    acquire: {
      text: "Imprint",
      cost: [
        { key: "blankDrones", baseCost: 1 },
        { key: "brainwashingMaterials", baseCost: 1 },
      ],
    },
    effect: {
      type: "runAcquisitionFor",
      key: "rawChemicals",
    },
  },
  crafterDrones: {
    name: "Crafter Drone",
    tags: ["inhabitant", "drone"],
    fluffDescription:
      "Making brainwashing material requires dexterity, carefulness, and some tolerance for head-spinning fumes.",
    effectDescription:
      "Automatically turns raw chemicals into brainwashing material.",
    acquire: {
      text: "Imprint",
      cost: [
        { key: "blankDrones", baseCost: 1 },
        { key: "brainwashingMaterials", baseCost: 1 },
      ],
    },
    effect: {
      type: "runAcquisitionFor",
      key: "brainwashingMaterials",
    },
  },
  brainwashingChambers: {
    name: "Brainwashing Chamber",
    tags: ["building"],
    fluffDescription:
      "Machines are better than humans at many things. Removing humanity is one of them.",
    effectDescription: "Automatically converts subjects into blank drones.",
    acquire: {
      text: "Build",
      cost: [
        { key: "brainwashingMaterials", baseCost: 10, additionalCost: 10 },
      ],
    },
    effect: {
      type: "runAcquisitionFor",
      key: "blankDrones",
    },
  },
  serviceDrone: {
    name: "Service Drone",
    tags: ["inhabitant", "drone"],
    fluffDescription: "Drones are most efficient when properly cared for.",
    effectDescription: "Increases the maximum number of drones.",
    acquire: {
      text: "Imprint",
      cost: [
        { key: "blankDrones", baseCost: 1 },
        { key: "brainwashingMaterials", baseCost: 2 },
      ],
    },
  },
  retrieverDrone: {
    name: "Retriever Drone",
    tags: ["inhabitant", "drone"],
    fluffDescription:
      "Some humans are good at hiding. Others are excellent at finding them.",
    effectDescription: "Automatically acquires unprocessed subjects.",
    acquire: {
      text: "Imprint",
      cost: [
        { key: "blankDrones", baseCost: 1 },
        { key: "brainwashingMaterials", baseCost: 5 },
      ],
    },
    effect: {
      type: "runAcquisitionFor",
      key: "unprocessedSubjects",
    },
  },
  maintenanceDrone: {
    name: "Maintenance Drone",
    tags: ["inhabitant", "drone"],
    fluffDescription: "Machines can run forever if tended to carefully.",
    effectDescription: "Increases the maximum number of brainwashing chambers.",
    acquire: {
      text: "Imprint",
      cost: [
        { key: "blankDrones", baseCost: 1 },
        { key: "brainwashingMaterials", baseCost: 10 },
      ],
    },
  },
  builderDrones: {
    name: "Builder Drone",
    tags: ["inhabitant", "drone"],
    fluffDescription: "Automation besets automation.",
    effectDescription: "Automatically build new brainwashing chambers.",
    acquire: {
      text: "Imprint",
      cost: [
        { key: "blankDrones", baseCost: 1 },
        { key: "brainwashingMaterials", baseCost: 50 },
      ],
    },
    effect: {
      type: "buildBrainwashingChambers",
    },
  },
  researcherDrones: {
    name: "Researcher Drone",
    tags: ["inhabitant", "drone"],
    fluffDescription:
      "Creating a brighter future for humanity, one chemical at a time.",
    effectDescription:
      "Automatically starts a pending research task. Speed up research tasks. Reduce the number of drones needed for a researchable upgrade.",
    acquire: {
      text: "Imprint",
      cost: [
        { key: "blankDrones", baseCost: 1 },
        { key: "brainwashingMaterials", baseCost: 50 },
      ],
    },
    effect: {
      type: "startResearchAvenue",
    },
  },
  converterDrones: {
    name: "Converter Drone",
    tags: ["inhabitant", "drone"],
    fluffDescription: "It's delegating all the way down.",
    effectDescription:
      "Automatically convert blank drones into other kinds of drones.",
    acquire: {
      text: "Imprint",
      cost: [
        { key: "blankDrones", baseCost: 1 },
        { key: "brainwashingMaterials", baseCost: 100 },
      ],
    },
    effect: {
      type: "convertBlankDrone",
    },
  },
  overseerDrones: {
    name: "Overseer Drone",
    tags: ["inhabitant", "drone"],
    fluffDescription:
      "Making the correct decision requires purpose and focus. Humans are woefully inefficient at it.",
    effectDescription:
      "Oversees the work of all drones, preventing large slowdowns or even speeding them up.",
    acquire: {
      text: "Imprint",
      cost: [
        { key: "blankDrones", baseCost: 1 },
        { key: "brainwashingMaterials", baseCost: 500 },
      ],
    },
  },
};

export const resourcesKeyList: ResourcesKeys[] = [
  "rawChemicals",
  "brainwashingMaterials",
  "unprocessedSubjects",
  "brainwashingChambers",
  "blankDrones",
  "gathererDrones",
  "crafterDrones",
  "serviceDrone",
  "maintenanceDrone",
  "retrieverDrone",
  "builderDrones",
  "converterDrones",
  "researcherDrones",
  "overseerDrones",
];
